import pickle
import uuid
from sqlalchemy import MetaData, Table, and_, update
from sqlalchemy.sql import column

from utsam.dbs.postgres import PgTable
from utsam.dbs.schemas.transactions.assignment_criteria_student_scores import PgAssignmentCriteriaStudentScoresTable
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.students import PgStudentsTable
from utsam.dbs.schemas.transactions.assignment_subcriteria_student_scores import PgAssignmentSubcriteriaStudentScoresTable
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable


class AssignmentCriteriaStudentScoresTable(PgTable):
    def __init__(
        self,
        criteria_json: dict,
        student_id: int,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        schema: str = "public",
        engine = None
    ):
        super().__init__(
            table_name = PgAssignmentCriteriaStudentScoresTable.name,
            user_name = user_name,
            password = password,
            host = host,
            db_name = db_name,
            port = port,
            engine = engine
        )
        self.criteria_json = criteria_json
        self.student_id = student_id
        self.criteria_id = None
        self.criteria_desc: str = None
        self.criteria_grade_desc = None
        self.weight = 0
        self.score = 0
        self.comments = None

        self.sql_statement = None
        self.where_dict = None
        self.where_filters = None
        self.extract_json()

    def extract_json(self):
        self.criteria_id = self.criteria_json.get('criteria_id')
        self.criteria_desc = self.criteria_json.get('criteria_desc')
        self.weight = self.criteria_json.get('weight')
        self.score = self.criteria_json.get('score')
        self.comments = self.criteria_json.get('comments')
        self.where_dict = {
            PgAssignmentCriteriasTable.criteria_id: self.criteria_id ,
            PgStudentsTable.student_id: self.student_id
        }
        self.where_filters = [column(key) == value for key, value in self.where_dict.items()]

    def generate_upsert_sql(self):
        if len(self.load_table(where_dict=self.where_dict)) > 0:
            self.sql_statement = self.generate_update_sql()
        else:
            self.sql_statement = self.generate_insert_sql()

    def generate_update_sql(self):
        return (
            self.table.update()
            .where(and_(*self.where_filters))
            .values(
                {
                    PgAssignmentCriteriaStudentScoresTable.criteria_student_score: self.score,
                    PgAssignmentCriteriaStudentScoresTable.criteria_student_score_comments: self.comments,
                }
            )
        )

    def generate_insert_sql(self):
        return (
            self.table.insert()
            .values(
                {
                    PgAssignmentCriteriaStudentScoresTable.criteria_student_score_id: str(uuid.uuid4()), 
                    PgAssignmentCriteriasTable.criteria_id: self.criteria_id, 
                    PgStudentsTable.student_id: self.student_id,
                    PgAssignmentCriteriaStudentScoresTable.criteria_student_score: self.score,
                    PgAssignmentCriteriaStudentScoresTable.criteria_student_score_comments: self.comments,
                }
            )
        )

    def upsert_marking(self):
        #self.create_marking_df()
        self.generate_upsert_sql()
        try:
            print(self.sql_statement)
            self.execute_sql(sql_statement=self.sql_statement)
        except Exception as e:
            print(f"Error when writing to database: {e}")
        

class AssignmentSubcriteriaStudentScoresTable(PgTable):
    def __init__(
        self,
        subcriteria_json: dict = None,
        student_id: int = None,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        schema: str = "public",
        engine = None
    ):
        super().__init__(
            table_name = PgAssignmentSubcriteriaStudentScoresTable.name,
            user_name = user_name,
            password = password,
            host = host,
            db_name = db_name,
            port = port,
            engine = engine
        )
        self.subcriteria_json = subcriteria_json
        self.student_id = student_id
        self.subcriteria_id = None
        self.subcriteria_desc: str = None
        self.subcriteria_grade_desc = None
        self.weight = 0
        self.score = 0
        self.comments = None
        self.sql_statement = None
        self.where_dict = None
        self.where_filters = None
        self.extract_json()

    def extract_json(self):
        self.subcriteria_id = self.subcriteria_json.get('subcriteria_id')
        self.subcriteria_desc = self.subcriteria_json.get('subcriteria_desc')
        self.weight = self.subcriteria_json.get('weight')
        self.score = self.subcriteria_json.get('score')
        self.comments = self.subcriteria_json.get('comments')
        self.where_dict = {
            PgAssignmentSubcriteriasTable.subcriteria_id: self.subcriteria_id,
            PgStudentsTable.student_id: self.student_id
        }
        self.where_filters = [column(key) == value for key, value in self.where_dict.items()]

    def generate_upsert_sql(self):
        if len(self.load_table(where_dict=self.where_dict)) > 0:
            self.sql_statement = self.generate_update_sql()
        else:
            self.sql_statement = self.generate_insert_sql()

    def generate_update_sql(self):
        return (
            self.table.update()
            .where(and_(*self.where_filters))
            .values(
                {
                    PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score: self.score,
                    PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score_comments: self.comments,
                }
            )
        )

    def generate_insert_sql(self):
        return (
            self.table.insert()
            .values(
                {
                    PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score_id: str(uuid.uuid4()), 
                    PgAssignmentSubcriteriasTable.subcriteria_id: self.subcriteria_id, 
                    PgStudentsTable.student_id: self.student_id,
                    PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score: self.score,
                    PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score_comments: self.comments,
                }
            )
        )

    def upsert_marking(self):
        #self.create_marking_df()
        self.generate_upsert_sql()
        try:
            print(self.sql_statement)
            self.execute_sql(sql_statement=self.sql_statement)
        except Exception as e:
            print(f"Error when writing to database: {e}")

        #try:
        #    for table in ['assignment_marking_criteria_student_scores_view', 'assignment_marking_subcriteria_student_scores_view', 'assignment_marking_detailed_student_scores_view']
        #    sql_statement = f"REFRESH MATERIALIZED VIEW {table};"
        #    print(sql_statement)
        #    self.execute_sql(sql_statement=sql_statement)
        #except Exception as e:
        #    print(f"Error when refreshing materialized view: {e}")


import pandas as pd

from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable
from utsam.bases.grades import get_grade


def get_criteria(data, criteria_id):
    criteria_mask = data[PgAssignmentCriteriasTable.criteria_id] == criteria_id
    print(data.loc[criteria_mask, PgAssignmentCriteriasTable.criteria_desc].drop_duplicates().values)
    return data.loc[criteria_mask, [PgAssignmentSubcriteriasTable.subcriteria_name, PgAssignmentSubcriteriasTable.subcriteria_desc, PgAssignmentSubcriteriasTable.subcriteria_weight]].drop_duplicates()


class Criteria:
    def __init__(
        self,
        criteria_id: str,
        criteria_desc: str = None,
    ):
        self.criteria_id: str = criteria_id
        self.criteria_desc: str = criteria_desc


class CriteriaMarking(Criteria):
    def __init__(
        self,
        criteria_id: str,
        criteria_df: pd.DataFrame,
        grades_df: pd.DataFrame,
    ):
        super().__init__(
            criteria_id = criteria_id,
        )
        self.criteria_df = criteria_df.copy()
        self.grades_df = grades_df.copy()
        self.criteria_grade_desc = None
        self.weight = 0
        self.subcriteria_dict: dict = {}
        self.score = 0
        self.grade = None
        self.comments = None
        self.set_criteria()

    def set_criteria(self):
        criteria_id_mask = self.criteria_df[PgAssignmentCriteriasTable.criteria_id].astype('str') == str(self.criteria_id)
        criteria_rows = self.criteria_df.loc[criteria_id_mask, ]
        if len(criteria_rows) > 0:
            criteria_rows.reset_index(inplace=True)
            self.criteria_desc = criteria_rows.at[0, PgAssignmentCriteriasTable.criteria_desc]
            self.weight = criteria_rows.at[0, PgAssignmentCriteriasTable.criteria_weight]
            self.criteria_grade_desc = criteria_rows[[
                PgAssignmentSubcriteriasTable.subcriteria_name,
                PgAssignmentSubcriteriasTable.subcriteria_desc,
                PgAssignmentSubcriteriasTable.subcriteria_weight,
                ]]
            self.criteria_grade_desc = self.criteria_grade_desc.drop_duplicates()
        else:
            print(f"No criteria found for {self.criteria_id}")

    def reset_score_and_grade(self):
        self.score = 0
        self.grade = None

    def set_score_and_grade(self):
        if len(self.subcriteria_dict) > 0:
            self.reset_score_and_grade()
            self.score = sum([subcrit.score * subcrit.weight for _, subcrit in self.subcriteria_dict.items()])
            self.grade = get_grade(self.grades_df, self.score)

    def set_comments(self):
        if len(self.subcriteria_dict) > 0:
            self.comments = '\n'.join([subcrit.comments for _, subcrit in self.subcriteria_dict.items()])

    def check_weights(self):
        if len(self.subcriteria_dict) > 0:
            assert sum([subcrit.weight for _, subcrit in self.subcriteria_dict.items()]) == 1, "Sum of weights doesn't equal to 1 (Missing subcriteria)"

    def print(self):
        print(f"Score for '{self.criteria_desc}' = {self.score} ({self.grade})")
        for subscrit_id, subcrit in self.subcriteria_dict.items():
            print(f"{subcrit.subcriteria_desc}: {subcrit.score} * {subcrit.weight} ({subcrit.grade})")

    def to_dict(self):
        if len(self.subcriteria_dict) > 0:
            return {
                "criteria_id": self.criteria_id,
                "criteria_desc": self.criteria_desc,
                "weight": self.weight,
                "grade": self.grade,
                "score": self.score,
                "comments": self.comments,
                "subcriterias": {subcriteria_id: subcrit.to_dict() for subcriteria_id, subcrit in self.subcriteria_dict.items()}
            }

    def get_criteria_grade_desc(self):
        print(self.criteria_desc)
        return self.criteria_grade_desc

    def append_subcriteria(self, subcriteria):
        self.subcriteria_dict[subcriteria.subcriteria_id] = subcriteria

    def set_marking_items(self):
        self.set_score_and_grade()
        self.set_comments()



from utsam.dbs.schemas.masterdata.assignments import PgAssignmentsTable
from utsam.dbs.schemas.masterdata.courses import PgCoursesTable
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.employees import PgEmployeesTable
from utsam.dbs.schemas.masterdata.grades import PgGradesTable
from utsam.dbs.schemas.masterdata.semesters import PgSemestersTable
from utsam.dbs.schemas.masterdata.sessions import PgSessionsTable
from utsam.dbs.schemas.masterdata.students import PgStudentsTable
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable



import os
import pickle 
from sqlalchemy import create_engine, MetaData, Table, and_
from sqlalchemy.sql import column
from pathlib import Path

from utsam.bases.params import _METADATA_


class PgDatabase:
    def __init__(
        self,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        echo = False,
        engine = None,
    ):
        self.database_type = 'postgresql'
        self.user_name = user_name
        self.password = password
        self.host = host
        self.db_name = db_name
        self.port = port
        self.connection_str = f'{self.database_type}://{self.user_name}:{self.password}@{self.host}:{self.port}/{self.db_name}'
        self.engine = engine
        self.create_engine()

    def execute_sql(self, sql_statement):
        with self.engine.connect() as conn:
            if sql_statement.is_select:
                return conn.execute(sql_statement).fetchall()
            conn.execute(sql_statement)
            conn.commit()

    def create_engine(self):
        if self.engine is None:
            self.engine = create_engine(self.connection_string, pool_size=10, max_overflow=0, echo=False)


class PgTable(PgDatabase):
    def __init__(
        self,
        table_name: str = None,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        schema: str = "public",
        engine = None
    ):
        super().__init__(
            user_name = user_name,
            password = password,
            host = host,
            db_name = db_name,
            port = port,
            engine = engine,
        )
        self.table_name= table_name
        self.schema = schema
        self.metadata = None
        self.table = None
        self.meta_path = Path.cwd() / _METADATA_
        self.meta_file_name = f'{self.table_name}.metadata'
        self.load_metadata()

    def load_metadata(self):
        if os.path.isfile(self.meta_path / self.meta_file_name):  
            self.reload_metadata()
        else:
            self.metadata = MetaData(schema=self.schema)
            self.metadata.reflect(self.engine, only=[self.table_name])
            self.table = Table(self.table_name, self.metadata, autoload=True, autoload_with=self.engine)

            temp_table = self.table.to_metadata(MetaData())
            with open(self.meta_path / self.meta_file_name, 'wb') as f:
                pickle.dump(temp_table, f)

    def reload_metadata(self):
        with open(self.meta_path / self.meta_file_name, 'rb') as f:
            self.table = pickle.load(f)

    def load_table(self, where_dict: dict = None):
        filters = [column(key) == value for key, value in where_dict.items()]
        stmt = (
            self.table.select().where(and_(*filters))
        )
        return self.execute_sql(sql_statement=stmt)


from dataclasses import dataclass

@dataclass
class PgAssignmentSubcriteriasTable:
    name: str = "assignment_subcriterias"
    subcriteria_id: str = 'subcriteria_id'
    subcriteria_name: str = 'subcriteria_name'
    subcriteria_desc: str = 'subcriteria_desc'
    subcriteria_weight: str = 'subcriteria_weight'


import pandas as pd

from utsam.bases.grades import get_score
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable
from utsam.dbs.schemas.masterdata.grades import PgGradesTable
from utsam.dbs.schemas.transactions.assignment_subcriteria_grade import PgAssignmentSubcriteriaGrade
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable


def get_subcriteria(data, criteria_id, subcriteria_id):
    criteria_mask = data[PgAssignmentCriteriasTable.criteria_id] == criteria_id
    subcriteria_mask = data[PgAssignmentSubcriteriasTable.subcriteria_id] == subcriteria_id
    print(data.loc[criteria_mask & subcriteria_mask, PgAssignmentSubcriteriasTable.subcriteria_desc].drop_duplicates().values)
    return data.loc[criteria_mask & subcriteria_mask, [PgGradesTable.grade_abbrev, PgAssignmentSubcriteriaGrade.subcriteria_grade_desc]].drop_duplicates()


class SubcriteriaMarking:
    def __init__(
        self,
        subcriteria_id: str,
        criteria_df: pd.DataFrame,
        grades_df: pd.DataFrame,
        grade: str = None,
        finetune: int = 0,
        comments: str = None,
    ):
        self.subcriteria_id = subcriteria_id
        self.criteria_df = criteria_df.copy()
        self.subcriteria_desc = None
        self.subcriteria_grade_desc = None
        self.grades_df = grades_df.copy()
        self.weight = 0
        self.grade = None
        self.finetune = None
        self.comments = None
        self.score = 0
        self.set_subcriteria()

    def set_subcriteria(self):
        subcriteria_id_mask = self.criteria_df[PgAssignmentSubcriteriasTable.subcriteria_id].astype('str') == str(self.subcriteria_id)
        subcriteria_rows = self.criteria_df.loc[subcriteria_id_mask, ]
        if len(subcriteria_rows) > 0:
            subcriteria_rows.reset_index(inplace=True)
            self.subcriteria_desc = subcriteria_rows.at[0, PgAssignmentSubcriteriasTable.subcriteria_desc]
            self.weight = subcriteria_rows.at[0, PgAssignmentSubcriteriasTable.subcriteria_weight]
            self.subcriteria_grade_desc = subcriteria_rows[[
                PgGradesTable.grade_abbrev,
                PgAssignmentSubcriteriaGrade.subcriteria_grade_desc,
                ]]
            self.subcriteria_grade_desc = self.subcriteria_grade_desc.drop_duplicates()
        else:
            print(f"No subcriteria found for {self.subcriteria_id}")

    def print(self):
        print(f"Score for '{self.subcriteria_desc}' = {self.score} ({self.grade})")

    def to_dict(self):
        return {
            "subcriteria_id": self.subcriteria_id,
            "subcriteria_desc": self.subcriteria_desc,
            "weight": self.weight,
            "grade": self.grade,
            "finetune": self.finetune,
            "comments": self.comments,
            "score": self.score,
        }

    def get_subcriteria_grade_desc(self):
        print(self.subcriteria_desc)
        return self.subcriteria_grade_desc

    def set_marking_items(self, grade, finetune, comments):
        self.grade = grade
        self.finetune = finetune
        self.comments = comments
        self.score = get_score(self.grades_df, self.grade, self.finetune)


from dataclasses import dataclass

@dataclass
class PgStudentsTable:
    name: str = "students"
    student_id: str = 'student_id'
    canvas_student_id: str = 'canvas_student_id'
    student_name: str = 'student_name'
    student_email: str = 'student_email'
    student_status: str = 'student_status'
    group_id: str = 'group_id'



from dataclasses import dataclass

@dataclass
class PgSemestersTable:
    name: str = "semesters"
    semester_id: str = 'semester_id'
    semester_name: str = 'semester_name'
    semester_abbrev2: str = 'semester_abbrev2'
    semester_abbrev3: str = 'semester_abbrev3'



from dataclasses import dataclass

@dataclass
class PgEmployeesTable:
    name: str = "employees"
    employee_id: str = 'employee_id'
    employee_first_name: str = 'employee_first_name'
    employee_last_name: str = 'employee_last_name'
    staff_id: str = 'staff_id'




from dataclasses import dataclass

@dataclass
class PgAssignmentCriteriasTable:
    name: str = "assignment_criterias"
    criteria_id: str = 'criteria_id'
    criteria_desc: str = 'criteria_desc'
    criteria_weight: str = 'criteria_weight'


import pandas as pd

from utsam.bases.params import ZSCORE_COL
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable
from utsam.dbs.schemas.transactions.assignment_criteria_student_scores import PgAssignmentCriteriaStudentScoresTable
from utsam.dbs.schemas.masterdata.assignments import PgAssignmentsTable
from utsam.dbs.schemas.masterdata.employees import PgEmployeesTable
from utsam.dbs.schemas.masterdata.students import PgStudentsTable


class AssignmentCriteriaMarkingTutor:
    def __init__(
        self,
        assignment_id: str,
        criteria_id: str,
        tutor_first_name: int,
        all_marks_df: pd.DataFrame,
        zscore_name: str = ZSCORE_COL
    ):
        self.assignment_id = assignment_id
        self.criteria_id = criteria_id
        self.tutor_first_name = tutor_first_name
        self.all_marks_df = all_marks_df.copy()
        self.zscore_name = zscore_name
        self.assignment_mask = None
        self.criteria_mask = None
        self.tutor_mask = None
        self.tutor_marks_df: pd.DataFrame = None
        self.tutor_mean: float = None
        self.tutor_std: float = None
        
    def filter_marks(self):
        self.assignment_mask = self.all_marks_df[PgAssignmentsTable.assignment_id].astype('str') == str(self.assignment_id)
        self.criteria_mask = self.all_marks_df[PgAssignmentCriteriasTable.criteria_id].astype('str') == str(self.criteria_id)
        self.tutor_mask = self.all_marks_df[PgEmployeesTable.employee_first_name] == self.tutor_first_name
        self.tutor_marks_df = self.all_marks_df[self.assignment_mask & self.criteria_mask & self.tutor_mask].copy()

    def get_students_ids(self):
        if len(self.tutor_marks_df) > 0:
            return list(set(self.tutor_marks_df[PgStudentsTable.student_id].sort_values().unique()))
        return []
        
    def compute_zscore(self):
        self.tutor_mean = self.tutor_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_score].mean()
        self.tutor_std = self.tutor_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_score].std(ddof=0)
        self.tutor_marks_df[self.zscore_name] = (self.tutor_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_score] - self.tutor_mean) / self.tutor_std

    def calculate_new_score(self, new_mean, new_std):
        self.tutor_marks_df[f"new_{PgAssignmentCriteriaStudentScoresTable.criteria_student_score}"] = self.tutor_marks_df[self.zscore_name] * new_std + new_mean
        self.tutor_marks_df[f"new_{PgAssignmentCriteriaStudentScoresTable.criteria_student_score}"].clip(lower=0, upper=100, inplace=True)

    def standardise(self, new_mean, new_std):
        self.filter_marks()
        self.compute_zscore()
        self.calculate_new_score(new_mean=new_mean, new_std=new_std)


from dataclasses import dataclass

@dataclass
class PgAssignmentsTable:
    name: str = "assignments"
    assignment_id: str = 'assignment_id'
    assignment_name: str = 'assignment_name'
    assignment_abbrev: str = 'assignment_abbrev'



from dataclasses import dataclass

@dataclass
class SemesterParams():
    spring: str = "SPR" 
    autumn: str = "AUT"

_SELECT_ = 'Select'

_CACHE_ = 'cache'
_LOGS_ = 'logs'
_METADATA_ = 'metadata'

ZSCORE_COL = 'zscore'


from dataclasses import dataclass

@dataclass
class PgAssignmentSubcriteriaStudentScoresTable:
    name: str = "assignment_subcriteria_student_scores"
    subcriteria_student_score_id: str = 'subcriteria_student_score_id'
    subcriteria_student_score: str = 'subcriteria_student_score'
    subcriteria_student_score_comments: str = 'subcriteria_student_score_comments'



from dataclasses import dataclass

@dataclass
class PgAssignmentCriteriaStudentScoresTable:
    name: str = "assignment_criteria_student_scores"
    criteria_student_score_id: str = 'criteria_student_score_id'
    criteria_student_score: str = 'criteria_student_score'
    criteria_student_score_comments: str = 'criteria_student_score_comments'

